
#Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


class TRT_MODEL:
    def __init__(self,engine_path,image_size,names_path,conf_thresh=0.2,nms_thresh=0.6,class_name = None,num_classes=2):

        
        self.TRT_LOGGER = trt.Logger(trt.Logger.INFO)
        self.IN_IMAGE_H, self.IN_IMAGE_W = image_size
        #Read Engine
        self.engine= self.get_engine(engine_path)
        # Create context
        self.context = self.engine.create_execution_context() 
        # Create Buffer
        self.buffers = self.allocate_buffers(1)
        #
        self.context.set_binding_shape(0, (1, 3, self.IN_IMAGE_H, self.IN_IMAGE_W))
        #
        self.class_name = class_name
        self.conf_thresh = conf_thresh
        self.nms_thresh = nms_thresh
        self.names_path = names_path
        self.num_classes = num_classes

    def get_engine(self,engine_path):
        # If a serialized engine exists, use it instead of building an engine.
        print("Reading engine from file {}".format(engine_path))
        with open(engine_path, "rb") as f, trt.Runtime(self.TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    
    def allocate_buffers(self,batch_size):
        inputs = []
        outputs = []
        bindings = []
        stream = cuda.Stream()
        for binding in self.engine:

            size = trt.volume(self.engine.get_binding_shape(binding)) * batch_size
            dims = self.engine.get_binding_shape(binding)
            print(size)
            # in case batch dimension is -1 (dynamic)
            if dims[0] < 0:
                size *= -1
            
            dtype = trt.nptype(self.engine.get_binding_dtype(binding))
            
            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            
            # Append the device buffer to device bindings.
            bindings.append(int(device_mem))
            # Append to the appropriate list.
            if self.engine.binding_is_input(binding):
                inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                outputs.append(HostDeviceMem(host_mem, device_mem))
        
        return inputs, outputs, bindings, stream

    def _preprocess_input(self,img):
        resized = cv2.resize(img, (self.IN_IMAGE_W, self.IN_IMAGE_H), interpolation=cv2.INTER_LINEAR)
        img_in = cv2.cvtColor(resized, cv2.COLOR_BGR2RGB)
        img_in = np.transpose(img_in, (2, 0, 1)).astype(np.float32)
        img_in = np.expand_dims(img_in, axis=0)
        img_in /= 255.0
        img_in = np.ascontiguousarray(img_in)
        # print("Shape of the network input: ", img_in.shape)
        return img_in

    def _post_processing(self,img,output):
        box_array = output[0]
        
        confs = output[1]
        
        if type(box_array).__name__ != 'ndarray':
            box_array = box_array.cpu().detach().numpy()
            confs = confs.cpu().detach().numpy()

        
        box_array = box_array[:, :, 0]
        
        
        max_conf = np.max(confs, axis=2)
        max_id = np.argmax(confs, axis=2)

        
        
        bboxes_batch = []
        for i in range(box_array.shape[0]):

            argwhere = max_conf[i] > self.conf_thresh
            l_box_array = box_array[i, argwhere, :]
            l_max_conf = max_conf[i, argwhere]
            l_max_id = max_id[i, argwhere]

            keep = self._nms(l_box_array, l_max_conf)

            bboxes = []
            if (keep.size > 0):
                l_box_array = l_box_array[keep, :]
                l_max_conf = l_max_conf[keep]
                l_max_id = l_max_id[keep]

                for j in range(l_box_array.shape[0]):
                    bboxes.append([l_box_array[j, 0], l_box_array[j, 1], l_box_array[j, 2], l_box_array[j, 3], l_max_conf[j], l_max_conf[j], l_max_id[j]])

            bboxes_batch.append(bboxes)
        
        
        return bboxes_batch


    def _load_class_names(self,namesfile):
        class_names = []
        with open(namesfile, 'r') as fp:
            lines = fp.readlines()
        for line in lines:
            line = line.rstrip()
            class_names.append(line)
        return class_names

    def _nms(self,boxes, confs,min_mode=False):
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]

        areas = (x2 - x1) * (y2 - y1)
        order = confs.argsort()[::-1]

        keep = []
        while order.size > 0:
            idx_self = order[0]
            idx_other = order[1:]

            keep.append(idx_self)

            xx1 = np.maximum(x1[idx_self], x1[idx_other])
            yy1 = np.maximum(y1[idx_self], y1[idx_other])
            xx2 = np.minimum(x2[idx_self], x2[idx_other])
            yy2 = np.minimum(y2[idx_self], y2[idx_other])

            w = np.maximum(0.0, xx2 - xx1)
            h = np.maximum(0.0, yy2 - yy1)
            inter = w * h

            if min_mode:
                over = inter / np.minimum(areas[order[0]], areas[order[1:]])
            else:
                over = inter / (areas[order[0]] + areas[order[1:]] - inter)

            inds = np.where(over <= self.nms_thresh)[0]
            order = order[inds + 1]

        return np.array(keep)


        # This function is generalized for multiple inputs/outputs.
        # inputs and outputs are expected to be lists of HostDeviceMem objects.
    def do_inference(self,img_in):
        inputs, outputs, bindings, stream = self.buffers
        inputs[0].host = img_in
        # Transfer input data to the GPU.
        [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
        # Run inference.
        self.context.execute_async(bindings=bindings, stream_handle=stream.handle)
        # Transfer predictions back from the GPU.
        [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
        # Synchronize the stream
        stream.synchronize()
        # Return only the host outputs.
        return [out.host for out in outputs]
    
    

    def detect_image(self,image_src):
        return_boxes = []
        img_in = self._preprocess_input(image_src)
        
        
        tic = time.time()
        trt_outputs = self.do_inference(img_in)
        toc = time.time()
        
        
        print('Inference time = ',(toc-tic))
        trt_outputs[0] = trt_outputs[0].reshape(1, -1, 1, 4)
        trt_outputs[1] = trt_outputs[1].reshape(1, -1, self.num_classes)
        
        
        
        
        boxes = self._post_processing(img_in, trt_outputs)
        
        class_names = self._load_class_names(self.names_path)
        width = image_src.shape[1]
        height = image_src.shape[0]
        
        for i in range(len(boxes[0])):
            box = boxes[0][i]
            x1 = int(box[0] * width)
            y1 = int(box[1] * height)
            x2 = int(box[2] * width)
            y2 = int(box[3] * height)
            class_id = box[6]
            return_boxes.append([class_names[class_id],x1,y1,x2,y2,])
        
        
        return return_boxes
        
        

#Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


class TRT_MODEL:
    def __init__(self,engine_path,image_size,names_path,conf_thresh=0.2,nms_thresh=0.6,class_name = None,num_classes=2):

        
        self.TRT_LOGGER = trt.Logger(trt.Logger.INFO)
        self.IN_IMAGE_H, self.IN_IMAGE_W = image_size
        #Read Engine
        self.engine= self.get_engine(engine_path)
        # Create context
        self.context = self.engine.create_execution_context() 
        # Create Buffer
        self.buffers = self.allocate_buffers(1)
        #
        self.context.set_binding_shape(0, (1, 3, self.IN_IMAGE_H, self.IN_IMAGE_W))
        #
        self.class_name = class_name
        self.conf_thresh = conf_thresh
        self.nms_thresh = nms_thresh
        self.names_path = names_path
        self.num_classes = num_classes

    def get_engine(self,engine_path):
        # If a serialized engine exists, use it instead of building an engine.
        print("Reading engine from file {}".format(engine_path))
        with open(engine_path, "rb") as f, trt.Runtime(self.TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    
    def allocate_buffers(self,batch_size):
        inputs = []
        outputs = []
        bindings = []
        stream = cuda.Stream()
        for binding in self.engine:

            size = trt.volume(self.engine.get_binding_shape(binding)) * batch_size
            dims = self.engine.get_binding_shape(binding)
            print(size)
            # in case batch dimension is -1 (dynamic)
            if dims[0] < 0:
                size *= -1
            
            dtype = trt.nptype(self.engine.get_binding_dtype(binding))
            
            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            
            # Append the device buffer to device bindings.
            bindings.append(int(device_mem))
            # Append to the appropriate list.
            if self.engine.binding_is_input(binding):
                inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                outputs.append(HostDeviceMem(host_mem, device_mem))
        
        return inputs, outputs, bindings, stream

    def _preprocess_input(self,img):
        resized = cv2.resize(img, (self.IN_IMAGE_W, self.IN_IMAGE_H), interpolation=cv2.INTER_LINEAR)
        img_in = cv2.cvtColor(resized, cv2.COLOR_BGR2RGB)
        img_in = np.transpose(img_in, (2, 0, 1)).astype(np.float32)
        img_in = np.expand_dims(img_in, axis=0)
        img_in /= 255.0
        img_in = np.ascontiguousarray(img_in)
        # print("Shape of the network input: ", img_in.shape)
        return img_in

    def _post_processing(self,img,output):
        box_array = output[0]
        
        confs = output[1]
        
        if type(box_array).__name__ != 'ndarray':
            box_array = box_array.cpu().detach().numpy()
            confs = confs.cpu().detach().numpy()

        
        box_array = box_array[:, :, 0]
        
        
        max_conf = np.max(confs, axis=2)
        max_id = np.argmax(confs, axis=2)

        
        
        bboxes_batch = []
        for i in range(box_array.shape[0]):

            argwhere = max_conf[i] > self.conf_thresh
            l_box_array = box_array[i, argwhere, :]
            l_max_conf = max_conf[i, argwhere]
            l_max_id = max_id[i, argwhere]

            keep = self._nms(l_box_array, l_max_conf)

            bboxes = []
            if (keep.size > 0):
                l_box_array = l_box_array[keep, :]
                l_max_conf = l_max_conf[keep]
                l_max_id = l_max_id[keep]

                for j in range(l_box_array.shape[0]):
                    bboxes.append([l_box_array[j, 0], l_box_array[j, 1], l_box_array[j, 2], l_box_array[j, 3], l_max_conf[j], l_max_conf[j], l_max_id[j]])

            bboxes_batch.append(bboxes)
        
        
        return bboxes_batch


    def _load_class_names(self,namesfile):
        class_names = []
        with open(namesfile, 'r') as fp:
            lines = fp.readlines()
        for line in lines:
            line = line.rstrip()
            class_names.append(line)
        return class_names

    def _nms(self,boxes, confs,min_mode=False):
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]

        areas = (x2 - x1) * (y2 - y1)
        order = confs.argsort()[::-1]

        keep = []
        while order.size > 0:
            idx_self = order[0]
            idx_other = order[1:]

            keep.append(idx_self)

            xx1 = np.maximum(x1[idx_self], x1[idx_other])
            yy1 = np.maximum(y1[idx_self], y1[idx_other])
            xx2 = np.minimum(x2[idx_self], x2[idx_other])
            yy2 = np.minimum(y2[idx_self], y2[idx_other])

            w = np.maximum(0.0, xx2 - xx1)
            h = np.maximum(0.0, yy2 - yy1)
            inter = w * h

            if min_mode:
                over = inter / np.minimum(areas[order[0]], areas[order[1:]])
            else:
                over = inter / (areas[order[0]] + areas[order[1:]] - inter)

            inds = np.where(over <= self.nms_thresh)[0]
            order = order[inds + 1]

        return np.array(keep)


        # This function is generalized for multiple inputs/outputs.
        # inputs and outputs are expected to be lists of HostDeviceMem objects.
    def do_inference(self,img_in):
        inputs, outputs, bindings, stream = self.buffers
        inputs[0].host = img_in
        # Transfer input data to the GPU.
        [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
        # Run inference.
        self.context.execute_async(bindings=bindings, stream_handle=stream.handle)
        # Transfer predictions back from the GPU.
        [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
        # Synchronize the stream
        stream.synchronize()
        # Return only the host outputs.
        return [out.host for out in outputs]
    
    

    def detect_image(self,image_src):
        return_boxes = []
        img_in = self._preprocess_input(image_src)
        
        
        tic = time.time()
        trt_outputs = self.do_inference(img_in)
        toc = time.time()
        
        
        print('Inference time = ',(toc-tic))
        trt_outputs[0] = trt_outputs[0].reshape(1, -1, 1, 4)
        trt_outputs[1] = trt_outputs[1].reshape(1, -1, self.num_classes)
        
        
        
        
        boxes = self._post_processing(img_in, trt_outputs)
        
        class_names = self._load_class_names(self.names_path)
        width = image_src.shape[1]
        height = image_src.shape[0]
        
        for i in range(len(boxes[0])):
            box = boxes[0][i]
            x1 = int(box[0] * width)
            y1 = int(box[1] * height)
            x2 = int(box[2] * width)
            y2 = int(box[3] * height)
            class_id = box[6]
            return_boxes.append([class_names[class_id],x1,y1,x2,y2,])
        
        
        return return_boxes
        
        
